
// PRODUCTOR
class Productor extends Thread {
    private Pila pila;
    private String alfabeto = "ABCDEFGHIJ#";

    public Productor( Pila letra){
	// copia del objeto compartido
        pila = letra;
    }
    public void run() {
        char letra = alfabeto.charAt(0); ;
        for( int i=1; letra!='#'; i++ ) //  Apila letras
        {   
            pila.apilar( letra );
            System.out.println("  "+letra);
            letra = alfabeto.charAt(i);
		    try {             //  Da una chance al hilo consumidor
                sleep( (int)(Math.random() * 200 ) );
            } catch( InterruptedException e ) {;}
        }                  //  for
		pila.FinPro();     //  Fin de la produccin 
    }
}

//  Notar que se crea una instancia de la clase Pila, y que 
// se utiliza el mtodo Pila.apilar() para ir produciendo.

// CONSUMIDOR
class Consumidor extends Thread {
    private Pila pila;
    public Consumidor( Pila t) {
    pila = t;
    }

    public void run() {
       char letra;           //  texto
       try {             //  Da una chance al hilo consumidor
           sleep(200 );
       } catch( InterruptedException e ) {;}
       while(pila.HayStock() || pila.HayProd()) 
          {    letra = pila.sacar();
               System.out.println("        "+letra);
               try {             //  Da una chance al hilo consumidor
                   sleep( (int)(Math.random() * 400 ) );
               } catch( InterruptedException e ) {;}
          } // while
    } // run()
}     // class Consumidor

//  En este caso, como en el del productor, se cuenta con un mtodo
// en la clase Pila, pila.recoger(), para consumir. Observe Ud.
// que nuestro consumidor dejar la pila impecablemente vacia ...

// MONITOR
class Pila {
    private char buffer[] = new char[10];
    private int siguiente = 0;
    // Sigen flags para saber el estado del buffer
    private boolean estaLlena = false;
    private boolean estaVacia = true;
    private boolean esperanza = true; // (De produccin)
    public  static void encabe() {System.out.println("Prod. Cons.");}
    public synchronized char sacar() {  // Mtodo para consumir
        while(estaVacia) //  Ciclo de espera por algo
        {                // para consumir; cuando haya
            try {        // saldremos  ... 
                wait(); 
            } catch( InterruptedException e ) {;}
        }
        //  No est vaca, entonces      
		siguiente--; // decrementamos, vamos consumir
        if( siguiente == 0 )  //  Fu la ltima, ya
            estaVacia = true; // no quedan mas letras
        estaLlena = false; // (Acabamos de consumir)
        notify();
        // Devuelve la letra al thread consumidor
        return( buffer[siguiente] );
    }

    public synchronized void apilar( char letra ) { // Mtodo para producir
        while( estaLlena == true ) //  Ciclo de espera por sitio donde
        {                          // almacenar esta nueva produccin.
            try {                  //  Lo habr cuando recoger() cambie  
                wait();            // esta bandera a false
            } catch( InterruptedException e ) {;}
       }
        buffer[siguiente++] = letra; // Incluimos en el buffer
        if( siguiente == 10 ) // Comprueba si el buffer est lleno
            estaLlena = true;
        estaVacia = false;
        notify();
    }
    public synchronized void FinPro() {     // Fin de la produccin 
		esperanza=false;
    }	
    public synchronized boolean HayProd() { // Hay an produccin ? 
    	return esperanza;
    }	
    public synchronized boolean HayStock() { // Pila con contenido ?
    	return !estaVacia;
    }	
}

/* En la clase Pila se pueden observar dos caractersticas importantes: los miembros dato (buffer[]) son privados, y los mtodos de acceso (apilar() y sacar()) son sincronizados.
 Aqu se observa que la variable estaVacia es un semforo.  La naturaleza privada de los datos evita que el productor y el consumidor accedan directamente a stos. Si se permitiese el acceso
  directo de ambos hilos de ejecucin a los datos, se podran producir problemas; por ejemplo, si el consumidor intenta retirar datos de un  buffer vaco, obtendr excepciones innecesarias, o se bloquear el proceso.  Los mtodos sincronizados de acceso impiden que los productores y 
 consumidores corrompan un objeto compartido. Mientras el productor est  apilando  una letra, el consumidor no la puede retirar y viceversa. (puede retirar otras). Est asegurada la integridad de cualquier objeto compartido.  El mtodo notify() al final de cada mtodo de acceso avisa a cualquier  proceso que est esperando por el objeto, entonces el proceso que ha estado esperando intentar acceder de nuevo al objeto. En el mtodo wait()  se hace que el hilo se quede a la espera de que le llegue un notify(), ya  sea enviado por el hilo de ejecucin o por el sistema.

 Ahora que ya se dispone de un productor, un consumidor y un objeto compartido,  se necesita una aplicacin que arranque los hilos y que consiga que todos  traten con el mismo objeto que estn compartiendo. Esto es lo que hace el  siguiente trozo de cdigo:                                   */

class ejemplo {
    public static void main( String args[] ) {
        Pila.encabe();   // Un mtodo esttico
        Pila pil = new Pila();
        Productor  pro = new Productor( pil );
        Consumidor con = new Consumidor( pil );
        pro.start();
        con.start();
    }
}
